package App::Greple::xlate::Lang;

use v5.14;
use warnings;

use Exporter 'import';
our @EXPORT_OK = qw(%LANGNAME);

our %LANGNAME = (
    'BG'    => 'Bulgarian',
    'CS'    => 'Czech',
    'DA'    => 'Danish',
    'DE'    => 'German',
    'EL'    => 'Greek',
    'EN'    => 'American English',
    'EN-GB' => 'British English',
    'EN-US' => 'American English',
    'ES'    => 'Spanish',
    'ET'    => 'Estonian',
    'FI'    => 'Finnish',
    'FR'    => 'French',
    'HU'    => 'Hungarian',
    'ID'    => 'Indonesian',
    'IT'    => 'Italian',
    'JA'    => 'Japanese',
    'KO'    => 'Korean',
    'LT'    => 'Lithuanian',
    'LV'    => 'Latvian',
    'NB'    => 'Norwegian',
    'NL'    => 'Dutch',
    'PL'    => 'Polish',
    'PT-BR' => 'Brazilian Portuguese',
    'PT-PT' => 'European Portuguese',
    'RO'    => 'Romanian',
    'RU'    => 'Russian',
    'SK'    => 'Slovak',
    'SL'    => 'Slovenian',
    'SV'    => 'Swedish',
    'TR'    => 'Turkish',
    'UK'    => 'Ukrainian',
    'ZH'    => 'Chinese',
    );

1;
